from math import exp, log, pi

import sympy

import identinum


def test_rational():
    sols = identinum.identify(3.0 / 7.0)
    assert sols[0] == sympy.Rational(3, 7)


def test_root2():
    sols = identinum.identify((3.0 / 7.0) ** 0.5)
    assert sols[0] == sympy.sqrt(sympy.Rational(3, 7))


def test_root3():
    sols = identinum.identify((3.0 / 7.0) ** (1.0 / 3.0))
    assert sols[0] == sympy.Rational(3, 7) ** sympy.Rational(1, 3)


def test_root34():
    sols = identinum.identify(3.0 ** 0.5 / 2.0)
    assert sols[0] == sympy.sqrt(sympy.Rational(3, 4))


def test_pi():
    sols = identinum.identify(3.0 / 4.0 * pi)
    assert sols[0] == sympy.pi * sympy.Rational(3, 4)


def test_sqrt_pi_2():
    sols = identinum.identify((0.5 * pi) ** 0.5)
    assert sols[0] == sympy.sqrt(sympy.pi * sympy.Rational(1, 2))


def test_exp_2_pi():
    sols = identinum.identify(exp(2 * pi))
    assert sols[0] == sympy.exp(sympy.pi * 2)


def test_logn_0_5():
    sols = identinum.identify(log(0.5))
    assert sols[0] == sympy.log(sympy.Rational(1, 2))


def test_pi2():
    sols = identinum.identify(pi ** 2 / 4.0)
    assert sols[0] == sympy.pi ** 2 * sympy.Rational(1, 4)


def test_poly():
    sols = identinum.findpoly(0.17852201277, tol=1.0e-10)
    assert sols == [-79852, -12268, 4735]


if __name__ == "__main__":
    test_rational()



import mpmath
from sympy import sympify


def identify(a, abs_tol=1.0e-15):
    # adding "pi" can sometimes hide desired solutions
    sols = mpmath.identify(a, tol=abs_tol, full=True)
    sols_pi = mpmath.identify(a, ["pi"], tol=abs_tol, full=True)
    lst = list({sympify(sol) for sol in sols + sols_pi})
    # sort by complexity
    num_ops = [expr.count_ops() for expr in lst]
    lst = [expr for _, expr in sorted(zip(num_ops, lst), key=lambda pair: pair[0])]
    return lst


def findpoly(a, tol=1.0e-15, maxcoeff=100000, max_degree=10):
    return mpmath.findpoly(a, n=max_degree, tol=tol, maxcoeff=maxcoeff)



import argparse

import sympy
from mpmath import mp

from .main import findpoly, identify


def main(argv=None):
    parser = argparse.ArgumentParser(
        description="Convert decimal to (approximate) rational number."
    )
    parser.add_argument("decimal", type=str, help="input decimal number")
    parser.add_argument(
        "--tolerance",
        "-t",
        type=float,
        help="absolute tolerance (default: precision of the given decimal)",
    )
    parser.add_argument(
        "--num-results",
        "-n",
        type=int,
        default=10,
        help="maximum number of results (default: 10)",
    )
    parser.add_argument(
        "--find-polysol",
        "-p",
        default=False,
        action="store_true",
        help=(
            "try to find a polynomial which has the input decimal"
            " as approximate solution (default: false)"
        ),
    )

    args = parser.parse_args(argv)

    if args.tolerance is None:
        num_digits = len(args.decimal.rsplit(".", 1)[-1])
        args.tolerance = 10 ** (-num_digits)

    # def _granularity(x, max_digits=100):
    #     '''Computes the "granularity" of a floating point number, i.e.,
    #     1.0e-3 for 3.141.
    #     '''
    #     for k in range(max_digits):
    #         if abs(x - int(x)) < 1.0e-15 * x:
    #             break
    #         x *= 10
    #     return 10**(-k)

    x = float(args.decimal)
    sols = identify(x, abs_tol=args.tolerance)
    if sols:
        if len(sols) > args.num_results:
            sols = sols[: args.num_results]
        errors = [sympy.N(float(args.decimal) - sol) for sol in sols]
        for sol, error in zip(sols, errors):
            print(f"{sol}   {error}")

    if args.find_polysol:
        poly_sol = findpoly(x, tol=args.tolerance * 10)
        if poly_sol is not None:
            mp.dps = 20
            residual = mp.polyval(poly_sol, x)
            n = len(poly_sol)
            print()
            print(
                f"{residual} = "
                + " + ".join(
                    ["{}*x^{}".format(a, n - k - 1) for k, a in enumerate(poly_sol)]
                )
            )

    if not sols and poly_sol is None:
        print("No expression found that approximates %r well enough." % args.decimal)



try:
    # Python 3.8+
    from importlib import metadata
except ImportError:
    try:
        import importlib_metadata as metadata
    except ImportError:
        __version__ = "unknown"

try:
    __version__ = metadata.version("identinum")
except Exception:
    __version__ = "unknown"
